# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Minimal subclassing as baselines
Adapted from https://github.com/pytorch/pytorch/tree/v1.11.0/benchmarks/overrides_benchmark
"""

from __future__ import annotations

import torch

__all__ = ["SubTensor", "SubWithTorchFunc"]


class SubTensor(torch.Tensor):
    pass


class SubWithTorchFunc(torch.Tensor):

    def __torch_function__(self, func, types, args=(), kwargs=None):
        return super().__torch_function__(func, types, args, {} if kwargs is None else kwargs)
